# -*- coding: utf-8 -*-
"""
Created on Sun Oct  6 23:10:40 2019

@author: patrick
"""
import numpy as np
import random
import time


class Game:
    
    clever_memory = [[0] * 19683, [0] * 19683]
    clever_threshold = 20
    monteCarlo_iterations = 20
    monteCarlo_depth = 3

    miniMax_memory = [[0] * 19683, [0] * 19683]
    miniMax_visited = [[0] * 19683, [0] * 19683]
    miniMax_isTableCreated = False
    

###############################################
#
#   Numbers for the squares
#
#   0 | 1 | 2
#   ---------
#   3 | 4 | 5
#   ---------
#   6 | 7 | 8
#
###############################################

        
    def play(self, strategy):
        Field = np.array([[0,0,0],[0,0,0],[0,0,0]])
        nextPlayer = random.randint(1, 2)
        won = 0

        memP1=[]
        memP2=[]

        while won == 0 and len(getChoicesLeft(Field)) > 0 :
            nextMove = self.pickNextMove(strategy, nextPlayer, Field)
            Field = placeMark(nextMove, Field, nextPlayer)
            rememberMove(Field, nextPlayer, memP1, memP2)
            nextPlayer = nextPlayer % 2 + 1
            won = haswon(Field)

        if strategy[0] == 'clever' or strategy[1] == 'clever':
            self.clever_store(won, memP1, memP2)

        return Field, won


    def pickNextMove(self, strategy, nextPlayer, Field):
        if strategy[nextPlayer - 1] == 'clever':
            return self.clever_pick(nextPlayer, Field)

        if strategy[nextPlayer - 1] == 'montecarlo':
            return self.monteCarlo_pick(nextPlayer, Field)

        if strategy[nextPlayer - 1] == 'minimax':
            self.miniMax_createTable()

            if len(getChoicesLeft(Field)) == 9:
                return random.choice([0, 2, 6, 8])
            return self.miniMax_pick(nextPlayer, Field)

        return self.random_pick(Field)


    def miniMax_createTable(self):
        if self.miniMax_isTableCreated:
            return

        startTime = time.time()

        self.miniMaxTable_EvaluateFinishedGames(1)
        self.miniMaxTable_EvaluateFinishedGames(2)

        for round in [8,7,6,5,4,3,2,1]:
            for player in [1,2]:
                self.miniMaxTable_EvaluateIteration(round, player)

        self.miniMax_isTableCreated = True
        end = time.time()
        print("Building MiniMax Table:", "{:.4f}".format(end - startTime), "s")


    def miniMaxTable_EvaluateFinishedGames(self, player):
        possibleFields = [[0, 0, 0, 0, 0, 0, 0, 0, 0]]
        for round in range(1, 10):
            possibleFields = self.miniMaxTable_EvaluateRound(possibleFields, round, player)
            player = player % 2 + 1


    def miniMaxTable_EvaluateRound(self, possibleFields, round, player):
        newPossibleFields = []
        for field in possibleFields:
            for choice in getChoicesLeft(field):
                fieldCopy = placeMark1DField(choice, field, player)
                fieldEncoded = ter2dec(fieldCopy)
                if self.miniMax_visited[player - 1][fieldEncoded] == 0:
                    if haswon(reshape(fieldCopy)) == 0:
                        newPossibleFields.append(fieldCopy)
                        self.miniMax_visited[player - 1][fieldEncoded] = round
                        if round == 9:
                            self.miniMax_memory[player - 1][fieldEncoded] = 0
                    else:
                        self.miniMax_memory[player - 1][fieldEncoded] = player
                        self.miniMax_visited[player - 1][fieldEncoded] = -1
        return newPossibleFields

    def miniMaxTable_EvaluateIteration(self, round, player):
        while (round in self.miniMax_visited[player - 1]):
            fieldEncoded = self.miniMax_visited[player - 1].index(round)
            Field = reshape(dec2ter(fieldEncoded))

            opponent = player % 2 + 1
            evaluate = self.miniMaxTable_EvaluateOpponent(opponent, Field)

            if opponent in evaluate:
                self.miniMax_memory[player - 1][fieldEncoded] = opponent
            elif 0 in evaluate:
                self.miniMax_memory[player - 1][fieldEncoded] = 0
            else:
                self.miniMax_memory[player - 1][fieldEncoded] = player
            self.miniMax_visited[player - 1][fieldEncoded] = -1


    def miniMaxTable_EvaluateOpponent(self, opponent, Field):
        evaluate = []
        for choice in getChoicesLeft(Field):
            encodedField = ter2dec(reshape(placeMark(choice, Field, opponent)))
            evaluate.append(self.miniMax_memory[opponent - 1][encodedField])
        return evaluate


    def clever_store(self, won, memP1, memP2):
        for mem in memP1:
            self.clever_memory[0][mem] += (won + 1) % 3 - 1
        for mem in memP2:
            self.clever_memory[1][mem] -= (won + 1) % 3 - 1


    def random_pick(self, Field):
        return random.choice(getChoicesLeft(Field))


    def miniMax_pick(self, nextPlayer, Field):
        win, draw = self.checkChoicesInMemory(Field, nextPlayer)

        if len(win) > 0:
            return random.choice(win)
        if len(draw) > 0:
            return random.choice(draw)
        return random.choice(Field)


    def checkChoicesInMemory(self, Field, nextPlayer):
        win = []
        draw = []
        for choice in getChoicesLeft(Field):
            FieldCopy = placeMark(choice, Field, nextPlayer)
            decodedField = ter2dec(reshape(FieldCopy))
            mem = self.miniMax_memory[nextPlayer - 1][decodedField]
            if mem == nextPlayer:
                win.append(choice)
            elif mem == 0:
                draw.append(choice)

        return win, draw


    def monteCarlo_pick(self, nextPlayer, Field):
        ratings = []

        for choice in getChoicesLeft(Field):
            fieldCopy = placeMark(choice, Field, nextPlayer)
            rating = self.monteCarlo_getRatingForChoice(nextPlayer, fieldCopy)
            ratings.append({'choice': choice, 'rating': rating})

        return monteCarlo_getBestChoice(ratings, nextPlayer)


    def monteCarlo_getRatingForChoice(self, nextPlayer, Field):
        rating = 0
        for iteration in range(0, self.monteCarlo_iterations):
            won = self.monteCarlo_getWinnerOfIteration(nextPlayer, Field)
            rating += (won + 1) % 3 - 1

        return rating


    def monteCarlo_getWinnerOfIteration(self, nextPlayer, Field):
        FieldCopy = Field.copy()
        counter = 0
        while len(getChoicesLeft(FieldCopy)) > 0 and counter < self.monteCarlo_depth:
            pick = self.random_pick(FieldCopy)
            FieldCopy = placeMark(pick, FieldCopy, nextPlayer)
            nextPlayer = nextPlayer % 2 + 1
            won = haswon(FieldCopy)
            counter += 1
            if won != 0:
                return won

        return 0


    def clever_pick(self, nextPlayer, Field):
        tryAll = self.clever_TryAllChoices(nextPlayer, Field)
        bestChoice = self.clever_SelectTheBestChoices(tryAll)
        self.clever_AddNoise(bestChoice)

        best = max(bestChoice, key=lambda x: x[1])
        return best[0]
        

    def clever_TryAllChoices(self, nextPlayer, Field):
        tryAll = []
        for choice in getChoicesLeft(Field):
            Test = Field.copy()
            field = np.reshape(Test, 9)
            field[choice] = nextPlayer
            decodedField = ter2dec(field)
            tryAll.append([choice, self.clever_memory[nextPlayer - 1][decodedField]])
        return tryAll


    def clever_SelectTheBestChoices(self, tryAll):
        bestChoices = []
        while len(tryAll) > 0:
            bestChoice = max(tryAll, key=lambda x: x[1])
            bestChoices.append(bestChoice)
            tryAll.remove(bestChoice)
            if bestChoices[0][1] > bestChoice[1] + self.clever_threshold:
                return bestChoices
        return bestChoices


    def clever_AddNoise(self, bestChoice):
        for choice in bestChoice:
            choice[1] += random.randint(0, self.clever_threshold)
        return bestChoice


def getChoicesLeft(field):
    if len(field) == 3:
        field = reshape(field)
    return [ind for ind, val in enumerate(field) if val == 0]


def placeMark(choice, Field, player):
    FieldCopy = Field.copy()
    row = choice // 3
    col = choice % 3
    if FieldCopy[row, col] != 0:
        raise Exception('The field' + str(choice) + 'is already taken by player' + str(FieldCopy[row, col]))
    FieldCopy[row, col] = player
    return FieldCopy


def placeMark1DField(choice, field, player):
    fieldCopy = field.copy()
    if fieldCopy[choice] != 0:
        raise Exception('The field' + str(choice) + 'is already taken by player' + str(fieldCopy(choice)))
    fieldCopy[choice] = player
    return fieldCopy


def rememberMove(Field, nextPlayer, memP1, memP2):
    field = np.reshape(Field, 9)
    fieldEncoded = ter2dec(field)
    if nextPlayer == 1:
        memP1.append(fieldEncoded)
    else:
        memP2.append(fieldEncoded)


def monteCarlo_getBestChoice(ratings, nextPlayer):
    if nextPlayer == 1:
        best = max(ratings, key=lambda x: x['rating'])
    else:
        best = min(ratings, key=lambda x: x['rating'])
    return best['choice']


def ter2dec(ter):
    dec = 0
    for idx in range(0, len(ter)):
        dec += ter[ -idx - 1] * 3 ** idx
    return dec
    

def dec2ter(dec):
    if dec == 0:
        return [0] * 9

    ter = []
    while dec != 0 :
        residual = dec % 3
        ter.insert(0, residual)
        dec = (dec - residual) // 3

    while len(ter) < 9:
        ter.insert(0,0)

    return ter
    

def reshape(Field):
    if len(Field) == 9:
        Matrix = np.zeros((3,3),dtype=int)
        Matrix[0,:] = Field[0:3]
        Matrix[1,:] = Field[3:6]
        Matrix[2,:] = Field[6:9]
        return Matrix
    elif len(Field) == 3:
        field = [0] * 9
        field[0:3] = Field[0,:]
        field[3:6] = Field[1,:]
        field[6:9] = Field[2,:]
        return field
    else:
        raise Exception('The input has invalid shape')
        

def haswon(Field):
    if (Field.prod(axis=0) == 1).any():
        return 1
    if (Field.prod(axis=0) == 8).any():
        return 2
    if (Field.prod(axis=1) == 1).any():
        return 1
    if (Field.prod(axis=1) == 8).any():
        return 2
    if Field[0, 0] * Field[1, 1] * Field[2, 2] == 1:
        return 1
    if Field[0, 0] * Field[1, 1] * Field[2, 2] == 8:
        return 2
    if Field[0, 2] * Field[1, 1] * Field[2, 0] == 1:
        return 1
    if Field[0, 2] * Field[1, 1] * Field[2, 0] == 8:
        return 2

    return 0
