import game
import numpy as np
import pytest

def test_haswon():
    F = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    assert game.haswon(F) == 0
    F = np.array([[1, 1, 1], [2, 0, 2], [0, 0, 0]])
    assert game.haswon(F) == 1
    F = np.array([[2, 0, 0], [2, 1, 1], [2, 0, 0]])
    assert game.haswon(F) == 2
    F = np.array([[0, 0, 1], [2, 0, 1], [0, 2, 1]])
    assert game.haswon(F) == 1
    F = np.array([[1, 2, 0], [0, 1, 2], [0, 0, 1]])
    assert game.haswon(F) == 1
    F = np.array([[0, 0, 2], [1, 2, 1], [2, 1, 0]])
    assert game.haswon(F) == 2


def test_dec2ter():
    assert game.dec2ter(0) == [0, 0, 0, 0, 0, 0, 0, 0, 0]
    assert game.dec2ter(2) == [0, 0, 0, 0, 0, 0, 0, 0, 2]
    assert game.dec2ter(10) == [0, 0, 0, 0, 0, 0, 1, 0, 1]
    assert game.dec2ter(100) == [0, 0, 0, 0, 1, 0, 2, 0, 1]


def test_ter2dec():
    assert game.ter2dec([0, 0, 0, 0, 0, 0, 0, 0, 0]) == 0
    assert game.ter2dec([0, 0, 0, 1, 0, 0, 0, 0, 0]) == 243
    assert game.ter2dec([0, 0, 0, 0, 0, 2, 0, 0, 0]) == 54
    assert game.ter2dec([0, 0, 0, 0, 0, 2, 0, 0, 1]) == 55
    assert game.ter2dec([0, 0, 0, 0, 0, 0, 1, 2, 1]) == 16


def test_placeMark():
    F = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    assert (game.placeMark(3, F, 1) == np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]])).all()
    F = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 0]])
    assert (game.placeMark(5, F, 2) == np.array([[0, 0, 0], [1, 0, 2], [0, 0, 0]])).all()
    F = np.array([[1, 0, 0], [0, 0, 0], [0, 0, 0]])
    with pytest.raises(Exception) as einfo:
        game.placeMark(0, F, 2)
    assert einfo.match('is already taken by player')


def test_getChoicesLeft():
    F = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    assert game.getChoicesLeft(F) == [0, 1, 2, 3, 4, 5, 6, 7, 8]
    F = np.array([[1, 2, 0], [0, 0, 0], [0, 2, 1]])
    assert game.getChoicesLeft(F) == [2, 3, 4, 5, 6]
    f = [1, 2, 0, 1, 0, 2, 0, 1, 0]
    assert game.getChoicesLeft(f) == [2, 4, 6, 8]
    f = [1, 2, 1, 1, 1, 2, 2, 1, 2]
    assert game.getChoicesLeft(f) == []


def test_monteCarlo_getWinnerOfIteration():
    simul = game.Game()
    Field = np.array([[1, 1, 0], [2, 0, 0], [2, 1, 2]])
    nextPlayer = 1
    assert simul.monteCarlo_getWinnerOfIteration(nextPlayer, Field) == 1

    Field = np.array([[2, 1, 2], [1, 2, 1], [0, 0, 0]])
    nextPlayer = 2
    assert simul.monteCarlo_getWinnerOfIteration(nextPlayer, Field) == 2

    Field = np.array([[2, 1, 2], [1, 2, 1], [0, 1, 0]])
    nextPlayer = 2
    assert simul.monteCarlo_getWinnerOfIteration(nextPlayer, Field) == 2

    Field = np.array([[2, 1, 2], [1, 2, 1], [1, 2, 0]])
    nextPlayer = 1
    assert simul.monteCarlo_getWinnerOfIteration(nextPlayer, Field) == 0


def test_getBestMonteCarloChoice():
    ratings = [
        {'choice': 1, 'rating': 8},
        {'choice': 4, 'rating': 15},
        {'choice': 5, 'rating': 1},
        {'choice': 8, 'rating': -5}
    ]
    assert game.monteCarlo_getBestChoice(ratings, 1) == 4
    assert game.monteCarlo_getBestChoice(ratings, 2) == 8
